////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2014-2022, Lawrence Livermore National Security, LLC.
// Produced at the Lawrence Livermore National Laboratory.
// Written by the LBANN Research Team (B. Van Essen, et al.) listed in
// the CONTRIBUTORS file. <lbann-dev@llnl.gov>
//
// LLNL-CODE-697807.
// All rights reserved.
//
// This file is part of LBANN: Livermore Big Artificial Neural Network
// Toolkit. For details, see http://software.llnl.gov/LBANN or
// https://github.com/LLNL/LBANN.
//
// Licensed under the Apache License, Version 2.0 (the "Licensee"); you
// may not use this file except in compliance with the License.  You may
// obtain a copy of the License at:
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
// implied. See the License for the specific language governing
// permissions and limitations under the license.
////////////////////////////////////////////////////////////////////////////////

#ifndef LBANN_LAYERS_IMAGE_BILINEAR_RESIZE_HPP_INCLUDED
#define LBANN_LAYERS_IMAGE_BILINEAR_RESIZE_HPP_INCLUDED

#include "lbann/layers/data_type_layer.hpp"
#include "lbann/layers/layer.hpp"

namespace lbann {

/** @brief Resize image with bilinear interpolation.
 *
 *  Tensors are assumed to be image data in CHW format. Gradients are
 *  not propagated during backprop.
 */
template <typename TensorDataType, data_layout Layout, El::Device Device>
class bilinear_resize_layer : public data_type_layer<TensorDataType> {
  static_assert(Layout == data_layout::DATA_PARALLEL,
                "bilinear_resize_layer only supports DATA_PARALLEL");
public:

  bilinear_resize_layer(lbann_comm *comm, El::Int height, El::Int width)
    : data_type_layer<TensorDataType>(comm), m_height(height), m_width(width) {
  }

  bilinear_resize_layer* copy() const override {
    return new bilinear_resize_layer(*this);
  }

  /** @name Serialization */
  ///@{

  template <typename ArchiveT>
  void serialize(ArchiveT& ar);

  ///@}

  std::string get_type() const override { return "bilinear resize"; }
  data_layout get_data_layout() const override { return Layout; }
  El::Device get_device_allocation() const override { return Device; }

  void fp_compute() override;

protected:

  friend class cereal::access;
  bilinear_resize_layer()
    : bilinear_resize_layer(nullptr, 1, 1)
  {}

  void setup_dims(DataReaderMetaData& dr_metadata) override {
    data_type_layer<TensorDataType>::setup_dims(dr_metadata);

    // Get input dimensions
    auto dims = this->get_input_dims();
    const auto& num_dims = dims.size();

    // Check that dimensions are valid
    std::stringstream err;
    if (num_dims < 2) {
      err << get_type() << " layer \"" << this->get_name() << "\" "
          << "expects input with at least two dimensions, "
          << "but input dimensions are ";
      for (size_t i = 0; i < num_dims; ++i) {
        err << (i > 0 ? " x " : "") << dims[i];
      }
      LBANN_ERROR(err.str());
    } else if (m_height <= 0) {
      err << get_type() << " layer \"" << this->get_name() << "\" "
          << "attempted to resize with "
          << "negative height (" << m_height << ")";
      LBANN_ERROR(err.str());
    } else if (m_width <= 0) {
      err << get_type() << " layer \"" << this->get_name() << "\" "
          << "attempted to resize with "
          << "negative width (" << m_width << ")";
      LBANN_ERROR(err.str());
    }

    // Resize output tensor
    dims[num_dims-2] = m_height;
    dims[num_dims-1] = m_width;
    this->set_output_dims(dims);

  }

private:

  /** Output image height.
   *  Data is assumed to be in CHW format.
   */
  El::Int m_height;
  /** Output image width.
   *  Data is assumed to be in CHW format.
   */
  El::Int m_width;
};

#ifndef LBANN_BILINEAR_RESIZE_LAYER_INSTANTIATE
#define PROTO_DEVICE(T, Device) \
  extern template class bilinear_resize_layer<T, data_layout::DATA_PARALLEL, Device>

#include "lbann/macros/instantiate_device.hpp"
#undef PROTO_DEVICE
#endif // LBANN_BILINEAR_RESIZE_LAYER_INSTANTIATE

} // namespace lbann

#endif // LBANN_LAYERS_IMAGE_BILINEAR_RESIZE_HPP_INCLUDED
